@@ -2,7 +2,7 @@ from __future__ import print_function
 
 import unittest
 import inspect
-import imp
+import types
 
 import wrapt
 
@@ -16,7 +16,7 @@ def passthru_decorator(wrapped, instance, args, kwargs):
     return wrapped(*args, **kwargs)
 """
 
-decorators = imp.new_module('decorators')
+decorators = types.ModuleType('decorators')
 exec_(DECORATORS_CODE, decorators.__dict__, decorators.__dict__)
 
 class OldClass1():
